import json
import urllib.request

from urllib.error import HTTPError

from grony.dotfile import Dotfile

from typing import Any, Dict


class Client:
    def __init__(self, dotfile: Dotfile) -> None:
        self.dotfile = dotfile

    def make_request(self, endpoint: str, **kwargs) -> Dict[str, Any]:
        port: int = self.dotfile.getint('config', 'ipc_port')
        secret: str = self.dotfile.get('config', 'secret')
        host: str = '127.0.0.1'
        data: bytes = urllib.parse.urlencode(kwargs).encode()

        req = urllib.request.Request(
            f'http://{host}:{port}/grony/{endpoint}',
            data=data, method='POST')
        req.add_header('Authorization', f"Bearer {secret}")

        try:
            response = urllib.request.urlopen(req)
            return json.loads(response.read().decode())
        except HTTPError as e:
            return json.loads(e.read())



import os
import re
import uuid
import logging
import configparser

from pathlib import Path

from typing import Any, Dict


class Dotfile(configparser.RawConfigParser):
    def __init__(self, path: str, *args, **kwargs) -> None:
        self.path = path
        super().__init__(*args, **kwargs)
        self.reload()

    def reload(self) -> None:
        self.clear()
        try:
            self.read(self.path)
        except Exception as ex:
            logging.warning(str(ex))

    def _get_repo_key(self, name: str) -> str:
        return f"repo '{name}'"

    def add_repo_section(self, name: str) -> configparser.SectionProxy:
        key = self._get_repo_key(name)
        self.add_section(key)
        return self[key]

    def get_repos(self) -> Dict[str, Dict[str, Any]]:
        return dict((match.group(1), self.get_repo(match.group(1)))
                    for match in (re.match(r"^\s*repo\s+'([^']+)'\s*$", s)
                                  for s in self.sections())
                    if match)

    def has_repo(self, name: str) -> bool:
        key = self._get_repo_key(name)
        return self.has_section(key)

    def get_repo(self, name: str) -> Dict[str, Any]:
        key = self._get_repo_key(name)
        if not self.has_section(key):
            return {}

        result = {'name': name}

        section = self[key]

        # Load .grony file, if any
        path = section.get('path', None)
        if path:
            conf_path = Path(os.path.join(path, '.grony'))
            if conf_path.exists() and conf_path.is_file():
                conf = configparser.RawConfigParser()
                conf.read(conf_path)
                if conf.has_section('repo'):
                    result.update(conf['repo'])

        # Settings in main conf override anything else
        result.update(section)

        return result

    def remove_repo(self, name: str) -> bool:
        key = self._get_repo_key(name)
        if not self.has_section(key):
            return False

        try:
            del self[key]
            save_dotfile(self)
            return True
        except Exception as ex:
            logging.warning(str(ex))
            return False


def _expand(path: str) -> str:
    return os.path.abspath(os.path.expandvars(path))


def _load_dotfile(path: str) -> Dotfile:
    return Dotfile(_expand(path))


def _set_default_value(dotfile: Dotfile, section: str,
                       key: str, value: str) -> bool:
    result = False
    if not dotfile.has_section(section):
        result = True
        dotfile.add_section(section)

    if not dotfile.has_option(section, key):
        result = True
        dotfile.set(section, key, value)

    return result


def add_defaults(dotfile: Dotfile) -> bool:
    """Adds the default config.
       Returns `True` if any defaul added. `False` otherwise.
    """

    defaults = {
        'ipc_port': '62830',
        'secret': str(uuid.uuid4()),
    }

    updated = False
    for k, v in defaults.items():
        updated = _set_default_value(dotfile, 'config', k, v) or updated

    if updated:
        logging.debug('Defaults added to dotfile.')

    return updated


def load_dotfile(path: str, with_defaults: bool = True) -> Dotfile:
    result = _load_dotfile(path)
    if with_defaults and add_defaults(result):
        save_dotfile(result)
        logging.info('Dotfile udpated with missing defaults.')
    return result


def save_dotfile(dotfile: Dotfile) -> None:
    logging.debug(f'Saving dotfile to {dotfile.path}...')
    
    try:
        os.makedirs(Path(dotfile.path).parent)
    except Exception as ex:
        logging.debug(f'While calling makedirs(): {ex}')

    try:
        with open(dotfile.path, "w+") as f:
            dotfile.write(f)
    except Exception as ex:
        logging.fatal(f"Can't update config at {dotfile.path}: {ex}")



import re
import cgi
import json
import logging
import urllib.parse

from functools import partial
from threading import Thread
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.error import HTTPError

from grony.dotfile import Dotfile, load_dotfile
from grony.commands import CallableCommand, Commands

from typing import Any, Dict, List, Optional


class ServerThread(Thread):
    def __init__(self, dotfile_path: str) -> None:
        super().__init__()
        dotfile = load_dotfile(dotfile_path)
        port = dotfile.getint('config', 'ipc_port')
        endpoint = ('127.0.0.1', port)
        self.server = HTTPServer(endpoint, partial(Handler, dotfile))

    def run(self):
        self.server.serve_forever()

    def stop(self):
        self.server.shutdown()


class Handler(BaseHTTPRequestHandler):
    def __init__(self, dotfile: Dotfile, *args, **kwargs) -> None:
        self.dotfile = dotfile
        super().__init__(*args, *kwargs)

    def send(self, data: Any, response_code: int = 200):
        self.send_response(response_code)
        self.send_header('Content-type', 'application/json')
        self.end_headers()
        self.wfile.write(json.dumps(data).encode())

    def _get_data(self) -> Dict[str, Any]:
        ctype, pdict = cgi.parse_header(self.headers.get('content-type', ''))
        if ctype == 'multipart/form-data':
            return cgi.parse_multipart(self.rfile, pdict)  # type: ignore
        elif ctype == 'application/x-www-form-urlencoded':
            length = int(self.headers.get('content-length', -1))
            qs = self.rfile.read(length).decode()
            return urllib.parse.parse_qs(qs, keep_blank_values=True)
        else:
            return {}

    def reject_request(self, message: str) -> None:
        logging.warning(message)
        self.send('Go home', 500)
        return

    def get_command(self) -> Optional[CallableCommand]:
        secret = self.dotfile.get('config', 'secret')

        if self.client_address[0] != '127.0.0.1':
            self.reject_request(
                f'Invalid request (address = {self.client_address})')
            return None

        auth_headers = (k for k, v in self.headers.items()
                        if k == "Authorization"
                        and v == f"Bearer {secret}")
        auth = next(auth_headers, None)
        if not auth:
            self.reject_request('Invalid request (authentication failed)')
            return None

        m = re.match(r'^\/grony\/([^\/]+)/?', self.path)
        if not m:
            self.reject_request('Invalid request')
            return None

        command = m.group(1)
        return Commands.get_command(command)

    def do_POST(self) -> None:
        fn = self.get_command()
        if not fn:
            self.reject_request("Unrecognized command")
            return

        try:
            messages: List[Dict[str, str]] = []

            # Convert from Dict[str, List[str]] to Dict[str, str], as
            # we only have a value for each key
            args = dict((k, v[0]) for k, v in self._get_data().items())

            success, msg = fn(self.dotfile, args)
            severity = 'success' if success else 'error'
            messages.append({'severity': severity, 'message': msg})
        except HTTPError as ex:
            logging.exception(ex)
            messages.append({'severity': 'fatal', 'message': str(ex)})
        finally:
            self.send({'messages': messages})



import os
import click
import signal
import logging

from pathlib import Path
from urllib.error import URLError

from grony.client import Client
from grony.server import ServerThread
from grony.dotfile import load_dotfile
from grony.scheduler import SchedulerThread
from grony.cli_output import info, success, err, warn, fatal

from tabulate import tabulate

from typing import Any, Dict, List, Optional

DEFAULT_CONF = os.environ.get('GRONY_CONFIG_PATH', None)
if not DEFAULT_CONF:
    dir = Path.joinpath(os.environ.get('XDG_CONFIG_HOME', Path.home()), '.grony')
    DEFAULT_CONF = Path.joinpath(dir, 'grony.conf')


def _is_success(result: Dict[str, Any]) -> bool:
    messages: Optional[List[Dict[str, str]]] = result.get('messages', None)
    if not messages:
        return False

    for m in messages:
        severity: str = m.get('severity', 'info')
        if severity.startswith('err'):
            return False

    return True


def _display_result(result: Dict[str, Any]) -> None:
    messages: List[Dict[str, str]] = result.get('messages', [])
    if not messages:
        fatal(f'Unknown error received from server: {str(result)}')
        return

    for msg in messages:
        severity: str = msg.get('severity', 'info')
        message = msg.get('message', '<Missing message>')

        fn = info if severity.startswith('info') \
            else success if severity.startswith('succ') \
            else warn if severity.startswith('warn') \
            else err

        fn(message)


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.option('--reload-delay', type=int, default=5, show_default=True,
              help='Delay between config reloads.')
@click.option('--log-level',
              type=click.Choice(['DEBUG', 'INFO', 'WARN', 'ERROR'],
                                case_sensitive=False),
              default='INFO', show_default=True, help='Log level.')
@click.option('--log-file', type=click.Path(file_okay=True),
              help='Logs to a file instead to stdout.')
@click.option('--dotfile', 'dotfile_path',
              type=click.Path(file_okay=True),
              default=DEFAULT_CONF,
              show_default=True, help='grony.conf location.')
def start(dotfile_path: str, reload_delay: int, log_level: str,
          log_file: Optional[str] = None) -> None:
    """Starts the main process.
    """

    logging.basicConfig(format='%(asctime)-5s %(levelname)-8s %(message)s',
                        filename=log_file,
                        level=getattr(logging, log_level.upper()))

    logging.info('===== Starting grony =====')

    scheduler_thread = SchedulerThread(dotfile_path, reload_delay)
    server_thread = ServerThread(dotfile_path)

    def handle_signal(sig: int, frame: Any) -> None:
        name = signal.Signals(sig).name
        logging.info(f'{name} received.')

        if scheduler_thread.is_alive():
            logging.info('Stopping scheduler process...')
            scheduler_thread.stop()

        if server_thread.is_alive():
            logging.info('Stopping server process...')
            server_thread.stop()

    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGTERM, handle_signal)

    logging.info('Press Ctrl+C to stop...')

    scheduler_thread.start()
    server_thread.start()

    scheduler_thread.join()
    server_thread.join()


@cli.command()
@click.argument('path', type=click.Path(exists=True, dir_okay=True),
                nargs=1)
@click.option('--name', help='Repository friendly name.')
@click.option('--dotfile', 'dotfile_path',
              type=click.Path(file_okay=True),
              default=DEFAULT_CONF,
              show_default=True, help='grony.conf location.')
def add(path: str, name: Optional[str], dotfile_path: str):
    """Adds a repository to the .grony.conf file.
    """

    dotfile = load_dotfile(dotfile_path)

    if not name:
        _, dirname = os.path.split(os.path.abspath(path))
        default: Optional[str] = None
        if dirname:
            default = dirname
        while not name or not name.strip():
            name = click.prompt("Repository friendly name", default=default)

    client = Client(dotfile)
    try:
        result = client.make_request('add', path=path, name=name)
        _display_result(result)
    except URLError as e:
        fatal(str(e.reason))


@cli.command()
@click.option('--dotfile', 'dotfile_path',
              type=click.Path(file_okay=True),
              default=DEFAULT_CONF,
              show_default=True, help='grony.conf location.')
@click.argument('name')
def remove(dotfile_path: str, name: str):
    """Removes a repository from the .grony.conf file.
    """

    dotfile = load_dotfile(dotfile_path)
    client = Client(dotfile)
    try:
        result = client.make_request('remove', name=name)
        _display_result(result)
    except URLError as e:
        fatal(str(e.reason))


@cli.command()
@click.option('--add', 'autoadd', is_flag=True, default=False,
              help="Also, add the repository to the default .grony.conf.")
@click.option('--dotfile', 'dotfile_path',
              type=click.Path(file_okay=True),
              default=DEFAULT_CONF,
              show_default=True, help='grony.conf location.')
@click.argument('path', type=click.Path(exists=True, dir_okay=True),
                default='.')
def init(autoadd: bool, dotfile_path: str, path: str):
    """Initializes a .grony file in the specified path.
    """

    dotfile = load_dotfile(dotfile_path)
    client = Client(dotfile)
    try:
        result = client.make_request('init', path=path)
        _display_result(result)
    except URLError as e:
        fatal(str(e.reason))

    if _is_success(result) and autoadd:
        add(path, None, dotfile_path)


@cli.command()
@click.option('--dotfile', 'dotfile_path',
              type=click.Path(file_okay=True),
              default=DEFAULT_CONF,
              show_default=True, help='grony.conf location.')
def list(dotfile_path: str):
    """List all configured repositories.
    """

    dotfile = load_dotfile(dotfile_path)
    items = tuple((name, repo.get('path'))
                  for (name, repo) in dotfile.get_repos().items())
    print(tabulate(items, headers=('Name', 'Path'), tablefmt='simple'))


@cli.command()
@click.option('--ini', 'ini_format', is_flag=True, default=False,
              help='Uses an output format suitable for config files')
@click.option('--dotfile', 'dotfile_path',
              type=click.Path(file_okay=True),
              default=DEFAULT_CONF,
              show_default=True, help='grony.conf location.')
@click.option('--all', 'show_all', is_flag=True,
              default=False,
              help='Display all repositories')
@click.argument('name', type=str, required=False)
def show(ini_format, dotfile_path: str, show_all: bool, name: str):
    """Show the effective settings for a repository.
    """

    dotfile = load_dotfile(dotfile_path)

    items = dotfile.get_repo(name).items()
    if ini_format:
        for k, v in items:
            print(f'{k} = {v}')
    else:
        print(tabulate(items, headers=('Key', 'Value'), tablefmt='simple'))




from setuptools import setup, find_packages
import os


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def main():
    setup(
        name='grony',
        version='0.4.2',
        description='An utility to schedule git-related actions using crontab expressions.',
        long_description=read('README.md'),
        long_description_content_type="text/markdown",
        keywords='git, automation, cron, crontab',

        author='Luis Medel',
        author_email='luis@luismedel.com',
        url='https://github.com/luismedel/grony',

        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Operating System :: Unix',
            'Topic :: Software Development :: Version Control',
            'Topic :: Software Development :: Version Control :: Git',
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)'
        ],

        package_dir={ '': 'src'},
        packages=find_packages('src'),
        include_package_data=True,
        install_requires=['click', 'tabulate', 'crontab'],
        entry_points={
            "console_scripts": [
                "grony=grony.main:main",
            ],
        },
    )

if __name__ == '__main__':
    main()




import click


def info(message: str) -> None:
    print(message)


def success(message: str) -> None:
    click.secho(message, fg='cyan')


def err(message: str) -> None:
    click.secho(f'Error: {message}', fg='red')


def warn(message: str) -> None:
    click.secho(f'Warning: {message}', fg='yellow')


def fatal(message: str, exit_code: int = 1) -> None:
    click.secho(f'Fatal: {message}', fg='red')
    exit(exit_code)
